/*
sudoku Version 0.1.0 - Sudoku for Curses
Copyright (C) 2005  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ncurses.h>
#include <time.h>
#include <sys/times.h>
#include <errno.h>
#include "rnd.h"

#define WAIT 100

typedef struct glbfmt {
   int run;          /* 0 = quit, 1 = run         */
   int reset;        /* 0 = no reset, 1 = reset   */
   int replay;       /* 0 = n, 1 = y same game    */
   int pause;        /* 0 = no pause, 1 = pause   */
   unsigned char *sd;     /* random number seed   */
   int col;          /* 0 - 8                     */
   int row;          /* 0 - 8                     */
   int num;          /* 1 - 9                     */
   int err;          /* Number of errors          */
   int totnum;       /* Number of filled squares  */
   time_t start;     /* start time in seconds     */
   time_t now;       /* curr  time in seconds     */
   int wait;         /* wait time in seconds      */
   int log;          /* handle for log file       */
   int card[96];     /* sudoku card 9x9           */
   int base[96];     /* sudoku base card 9x9      */
   } glbfmt;

void putstx(char *pgm)
   {
   fprintf(stderr,"Usage: %s\n", pgm);
   exit(1);
   } /* putstx */

void opn(glbfmt *glb)
   {
   glb->log = open("sudoku.log",
      O_WRONLY|O_CREAT|O_TRUNC,
      S_IRUSR|S_IWUSR);
   if (glb->log < 0)
      {
      fprintf(stderr,"sudoku.log\n");
      perror("Open error");
      exit(1);
      } /* if open error */
   } /* opn */

void putlog(glbfmt *glb, char *str)
   {
   int len;
   int wrtlen;
   len = strlen(str);
   wrtlen = write(glb->log,str,len);
   if (wrtlen != len)
      {
      fprintf(stderr,"sudoku.log\n");
      perror("Write error");
      exit(1);
      } /* if write error */
   } /* putlog */

void cls(glbfmt *glb)
   {
   int rslt;
   rslt = close(glb->log);
   if (rslt < 0)
      {
      fprintf(stderr,"sudoku.log\n");
      perror("Close error");
      exit(1);
      } /* if close error */
   } /* cls */

void putcard(glbfmt *glb)
   {
   int i,j;
   int *p;
   time_t now;
   struct tm *t;
   char *tstr;
   char cell[16];
   char str[128];
   time(&now);
   t = (struct tm *) localtime(&now);
   tstr = (char *) asctime(t);
   putlog(glb,tstr);
   putlog(glb,"\n");
   strcpy(str,"            Sudoku\n");
   putlog(glb,str);
   putlog(glb,"+---------+---------+---------+\n");
   p = (int *) glb->card;
   j = 9;
   while (j--)
      {
      strcpy(str,"|");
      i = 9;
      while (i--)
	 {
	 if (*p > 0)
	    sprintf(cell," %d ", *p);
	 else
	    sprintf(cell," - ");
	 strcat(str,cell);
         if (i == 6 || i == 3)
            {
            strcat(str,"|");
            } /* for each trio of columns */
	 p++;
	 } /* for each col in row */
      strcat(str,"|\n");
      putlog(glb,str);
      if (j == 6 || j == 3 || j == 0)
         {
         putlog(glb,"+---------+---------+---------+\n");
         } /* for each trio of lines */
      } /* for each row */
   } /* putcard */

void fillgroup(glbfmt *glb,
   int nums,
   int groupcol,
   int grouprow)
   {
   int i;
   int sqofst;
   int urnofst;
   int sq;
   int row;
   int col;
   int lastsq;
   int lasturn;
   int newnum;
   int badnum;
   int *p,*q;
   int *cardptr;
   int *sqptr;
   int *sqend;
   int sqlst[16];
   int *urnptr;
   int *urnend;
   int urnlst[16];
   p = (int *) sqlst;
   q = (int *) sqlst + 9;
   i = 0;
   while (p < q) *p++ = i++;
   p = (int *) urnlst;
   q = (int *) urnlst + 9;
   i = 1;
   while (p < q) *p++ = i++;
   lastsq = lasturn = 9;
   sqend  = (int *) sqlst + 8;     /* 9 - 1 */
   urnend = (int *) urnlst + 8;    /* 9 - 1 */
   i = nums;
   while (i-- && lasturn >= 0 && lastsq > 9 - nums)
      {
      sqofst = (int) rndnum(lastsq--,glb->sd);
      sqptr  = (int *) sqlst + sqofst;
      sq     = *sqptr;
      *sqptr = *sqend;
      *sqend-- = 999999999;
      col = groupcol + (sq % 3);
      row = grouprow + (sq / 3);
      newnum  = 0;
      while (lasturn >= 0)
	 {
	 urnofst = (int) rndnum(lasturn--,glb->sd);
         urnptr  = (int *) urnlst + urnofst;
	 newnum  = *urnptr;
         *urnptr = *urnend;
         *urnend-- = 999999999;
	 badnum = FALSE;
	 /* check for duplicate in same row */
	 p = (int *) glb->card + (row * 9);
	 q = (int *) p + 9;
	 while (p < q)
	    {
	    if (*p == newnum)
	       {
	       badnum = TRUE;
	       break;
	       } /* if redundant number */
	    p++;
	    } /* for each col in row */
	 if (badnum) continue;
	 /* check for duplicate in same col */
	 p = (int *) glb->card + col;
	 q = (int *) glb->card + 72;   /* 81 - 9 */
	 while (p < q)
	    {
	    if (*p == newnum)
	       {
	       badnum = TRUE;
	       break;
	       } /* if redundant number */
	    p += 9;
	    } /* for each row in col */
	 if (!badnum) break;
	 } /* ck if good number */
      if (lasturn < 0)
	 {
	 break;
	 endwin();
	 fprintf(stderr,"newnum %d "
	    "lasturn %d "
	    "col %d row %d\n",
	    newnum, lasturn,
	    col, row);
	 fprintf(stderr,"sq %d lastsq %d "
	    "groupcol %d grouprow %d\n",
	    sq, lastsq, groupcol, grouprow);
	 exit(0);
	 } /* if out of numbers */
      cardptr  = (int *) glb->card + (row * 9) + col;
      *cardptr = newnum;
      glb->totnum++;
      } /* for each fill_in */
   } /* fillgroup */

void initglb(glbfmt *glb)
   {
   int *p,*q,*r;
   time_t now;
   time(&now);
   glb->start  = now;   /* starting time */
   glb->run    = 1;
   glb->reset  = 0;
   glb->replay = 0;
   glb->pause  = 0;
   glb->col    = 0;
   glb->row    = 0;
   glb->num    = 0;
   glb->err    = 0;
   glb->totnum = 0;
   glb->wait   = WAIT;
   glb->sd = (unsigned char *) rndinit();
   if (glb->sd == NULL)
      {
      fprintf(stderr,"sudoku: out of memory "
	 "allocating glb->sd\n");
      exit(1);
      } /* if out of mem */
   opn(glb);
   p = (int *) glb->card;
   q = (int *) glb->card + 81;
   while (p < q) *p++ = 0;
   /* nums, col, row */
   fillgroup(glb,3,0,0);
   fillgroup(glb,4,3,0);
   fillgroup(glb,3,6,0);
   fillgroup(glb,3,0,3);
   fillgroup(glb,4,3,3);
   fillgroup(glb,3,6,3);
   fillgroup(glb,3,0,6);
   fillgroup(glb,4,3,6);
   fillgroup(glb,3,6,6);
   putcard(glb);
   p = (int *) glb->card;
   q = (int *) glb->card + 81;
   r = (int *) glb->base;
   while (p < q) *r++ = *p++;
   } /* initglb */

void initreplay(glbfmt *glb)
   {
   int *p,*q,*r;
   time_t now;
   time(&now);
   glb->start  = now;   /* restart time */
   glb->run    = 1;
   glb->reset  = 0;
   glb->replay = 0;
   glb->pause  = 0;
   glb->col    = 0;
   glb->row    = 0;
   glb->num    = 0;
   glb->err    = 0;
   glb->totnum = 0;
   glb->wait   = WAIT;
   p = (int *) glb->base;
   q = (int *) glb->base + 81;
   r = (int *) glb->card;
   while (p < q)
      {
      if (*p > 0) glb->totnum++;
      *r++ = *p++;
      } /* for each number on card */
   putcard(glb);
   } /* initreplay */

void initnc()
   {
   initscr();
   cbreak();
   noecho();
   nonl();
   intrflush(stdscr,FALSE);
   keypad(stdscr,TRUE);
   nodelay(stdscr,TRUE);
   curs_set(0);
   clear();
   refresh();
   } /* initnc */

void escape(glbfmt *glb)
   {
   int i;
   int ch;
   int *p;
   ch = getch();
   if (ch == 0x1b || ch == 'q')
      {
      glb->run = 0;
      } /* if quit */
   else if (ch == KEY_DOWN)
      {
      glb->row++;
      if (glb->row > 8)
	 glb->row = 0;
      p = (int *) glb->base + (glb->row * 9) + glb->col;
      i = 0;
      while (*p)
	 {
	 i++;
	 if (i > 9) break;
         glb->row++;
         if (glb->row > 8)
	    glb->row = 0;
	 p = (int *) glb->base + (glb->row * 9) + glb->col;
	 } /* for each original number */
      } /* if down cursor */
   else if (ch == KEY_UP)
      {
      glb->row--;
      if (glb->row < 0)
	 glb->row = 8;
      p = (int *) glb->base + (glb->row * 9) + glb->col;
      i = 0;
      while (*p)
	 {
	 i++;
	 if (i > 9) break;
         glb->row--;
         if (glb->row < 0)
	    glb->row = 8;
	 p = (int *) glb->base + (glb->row * 9) + glb->col;
	 } /* for each original number */
      } /* if up cursor */
   else if (ch == KEY_RIGHT)
      {
      glb->col++;
      if (glb->col > 8)
	 glb->col = 0;
      p = (int *) glb->base + (glb->row * 9) + glb->col;
      i = 0;
      while (*p)
	 {
	 i++;
	 if (i > 9) break;
         glb->col++;
         if (glb->col > 8)
	    glb->col = 0;
	 p = (int *) glb->base + (glb->row * 9) + glb->col;
	 } /* for each original number */
      } /* if right cursor */
   else if (ch == KEY_LEFT)
      {
      glb->col--;
      if (glb->col < 0)
	 glb->col = 8;
      p = (int *) glb->base + (glb->row * 9) + glb->col;
      i = 0;
      while (*p)
	 {
	 i++;
	 if (i > 9) break;
         glb->col--;
         if (glb->col < 0)
	    glb->col = 8;
	 p = (int *) glb->base + (glb->row * 9) + glb->col;
	 } /* for each original number */
      } /* if left cursor */
   else if (ch >= '1' && ch <= '9')
      {
      glb->num = ch - '0';
      } /* if valid number */
   else if (ch == 'b')
      {
      glb->replay = 1;
      } /* if replay */
   else if (ch == 'r')
      {
      glb->reset = 1;
      } /* if reset */
   else if (ch == 'p')
      {
      glb->pause ^= 1;
      } /* if pause */
   } /* escape */

void pawz(glbfmt *glb)
   {
   /* toggle 'p' */
   while (!glb->pause)
      {
      escape(glb);
      if (!glb->run)
         {
         endwin();
         cls(glb);
         exit(0);
         } /* if quit */
      } /* for each poll of kb */
   glb->pause = 0;
   } /* pawz */

void bldmenu(glbfmt *glb)
   {
   move(0,30);
   addstr("Sudoku");
   move(2,2);
   addstr("Down  = Next row");
   move(3,2);
   addstr("Up    = Previous row");
   move(4,2);
   addstr("Right = Next column");
   move(5,2);
   addstr("Left  = Previous column");
   move(6,2);
   addstr("b     = Replay same game");
   move(7,2);
   addstr("r     = Reset, new  game");
   move(8,2);
   addstr("q     = Quit");
   refresh();
   } /* bldmenu */

void bldscr(glbfmt *glb)
   {
   int i,j;
   int row;
   int col;
   int min;
   int sec;
   int *p;
   time_t now;
   int elapse;
   char cell[16];
   char str[128];
   time(&now);
   elapse = (int) now - glb->start;
   sec = elapse % 60;
   min = elapse / 60;
   sprintf(str,"Elapsed time: %3d:%02d",
      min, sec);
   move(10,2);
   clrtoeol();
   addstr(str);
   sprintf(str,"Errors: %12d", glb->err);
   move(11,2);
   clrtoeol();
   addstr(str);
   row = 2;
   move(row,35);
   clrtoeol();
   addstr("+---------+---------+---------+");
   row++;
   p = (int *) glb->card;
   j = 0;
   while (j < 9)
      {
      col = 35;
      move(row,col);
      clrtoeol();
      addstr("|");
      i = 0;
      while (i < 9)
	 {
	 if (*p > 0)
	    sprintf(cell," %d ", *p);
	 else
	    sprintf(cell," - ");
	 if (j == glb->row
	    && i == glb->col)
            attron(A_REVERSE);
	 addstr(cell);
         attroff(A_REVERSE);
	 i++;
         if (i == 3 || i == 6)
            {
            addstr("|");
            } /* for each trio of columns */
	 p++;
	 } /* for each col in row */
      addstr("|");
      j++;
      if (j % 3 == 0)
         {
         move(++row,35);
         clrtoeol();
         addstr("+---------+---------+---------+");
         } /* for each trio of lines */
      row++;
      } /* for each row */
   refresh();
   } /* bldscr */

void slp(glbfmt *glb)
   {
   int rslt;
   int lmt;
   char str[32];
   struct timespec request;
   struct timespec remain;
   lmt = glb->wait;
   request.tv_sec  = 0;
   request.tv_nsec = 1;
   remain.tv_sec   = 0;
   remain.tv_nsec  = 0;
   while (lmt--)
      {
      rslt = nanosleep(&request,&remain);
      if (rslt < 0)
	 {
	 endwin();
	 if (errno == EINTR)
	    strcpy(str,"EINTR");
	 else if (errno == EINVAL)
	    strcpy(str,"EINVAL");
	 else
	    strcpy(str,"unknown value");
	 fprintf(stderr,"req sec %u "
	    " req nsec %u "
	    " rem sec %u "
	    " rem nsec %u\n",
            (unsigned int) request.tv_sec,
            (unsigned int) request.tv_nsec,
	    (unsigned int) remain.tv_sec,
	    (unsigned int) remain.tv_nsec);
	 fprintf(stderr,"sudoku: slp: error %d %s\n",
	    errno, str);
	 perror("Nanosleep error");
	 exit(1);
	 } /* if err */
      } /* wait loop */
   } /* slp */

void addnum(glbfmt *glb)
   {
   int i,j;
   int groupcol;
   int grouprow;
   int newnum;
   int badnum;
   int *p,*q;
   int *currpos;
   int *cardptr;
   groupcol = (glb->col / 3) * 3;
   grouprow = (glb->row / 3) * 3;
   newnum  = glb->num;
   badnum = FALSE;
   /* check for duplicate in same row */
   p = (int *) glb->card + (glb->row * 9);
   q = (int *) p + 9;
   currpos = (int *) p + glb->col;
   while (p < q)
      {
      if (*p == newnum
	 && p != currpos)
         {
         badnum = TRUE;
         break;
         } /* if redundant number */
      p++;
      } /* for each col in row */
   if (badnum)
      {
      move(20,20);
      clrtoeol();
      addstr("Duplicate number in row");
      move(21,20);
      clrtoeol();
      addstr("Change number to continue");
      refresh();
      glb->err++;
      glb->num = 0;
      return;
      } /* if duplicate in row */
   /* check for duplicate in same col */
   p = (int *) glb->card + glb->col;
   q = (int *) glb->card + 72;   /* 81 - 9 */
   currpos = (int *) p + (glb->row * 9);
   while (p < q)
      {
      if (*p == newnum
	 && p != currpos)
         {
         badnum = TRUE;
         break;
         } /* if redundant number */
      p += 9;
      } /* for each row in col */
   if (badnum)
      {
      move(20,20);
      clrtoeol();
      addstr("Duplicate number in column");
      move(21,20);
      clrtoeol();
      addstr("Change number to continue");
      refresh();
      glb->err++;
      glb->num = 0;
      return;
      } /* if duplicate in col */
   p = (int *) glb->card + (grouprow * 9) + groupcol;
   currpos = (int *) glb->card + (glb->row * 9) + glb->col;
   j = 3;
   while (j--)
      {
      i = 3;
      while (i--)
         {
	 if (newnum == *p
	    && p != currpos)
	    {
	    badnum = TRUE;
	    break;
	    } /* if duplicate in group */
	 p++;
         } /* for each col in group */
      p += 6;
      } /* for each row in group */
   if (badnum)
      {
      move(20,20);
      clrtoeol();
      addstr("Duplicate number in group of 9");
      move(21,20);
      clrtoeol();
      addstr("Change number to continue");
      refresh();
      glb->err++;
      glb->num = 0;
      return;
      } /* if duplicate in row */
   move(20,20);
   clrtoeol();
   addstr("You added a valid number");
   move(21,20);
   clrtoeol();
   addstr("Move cursor to continue");
   refresh();
   cardptr  = (int *) glb->card + (glb->row * 9) + glb->col;
   *cardptr = newnum;
   glb->totnum++;
   glb->num = 0;
   } /* addnum */

void play(glbfmt *glb)
   {
   int *p;
   int indx;
   int min;
   int sec;
   int cardnum;
   int basenum;
   int totnum;
   int prevcol;
   int prevrow;
   time_t now;
   int elapse;
   char str[96];
   totnum = 0;
   glb->num = 0;
   prevcol = glb->col;
   prevrow = glb->row;
   while (glb->run
      && !glb->reset
      && !glb->replay
      && totnum < 81)
      {
      escape(glb);
      if (glb->col != prevcol
         || glb->row != prevrow)
	 {
         move(20,20);
         clrtoeol();
         move(21,20);
         clrtoeol();
         prevcol = glb->col;
         prevrow = glb->row;
	 } /* if cursor changed */
      bldscr(glb);
      indx = (glb->row * 9) + glb->col;
      p = (int *) glb->card + indx;
      cardnum = *p;
      p = (int *) glb->base + indx;
      basenum = *p;
      if (cardnum > 0
	 && cardnum == basenum)
	 {
	 move(20,20);
	 clrtoeol();
	 addstr("Cursor is on an original number");
	 move(21,20);
	 clrtoeol();
	 addstr("Move cursor and continue playing");
	 refresh();
	 slp(glb);
	 continue;
	 } /* wait for valid position */
      if (!glb->num)
	 {
	 slp(glb);
	 continue;
	 } /* wait for number */
      sprintf(str,"You have entered %d", glb->num);
      move(20,20);
      clrtoeol();
      addstr(str);
      move(21,20);
      clrtoeol();
      refresh();
      addnum(glb);
      slp(glb);
      prevcol = glb->col;
      prevrow = glb->row;
      } /* play until card filled */
   putlog(glb,"\n");
   putcard(glb);
   putlog(glb,"\n");
   time(&now);
   elapse = (int) now - glb->start;
   sec = elapse % 60;
   min = elapse / 60;
   sprintf(str,"Elapsed time: %4d:%02d\n",
      min, sec);
   putlog(glb,str);
   sprintf(str,"Total numbers: %6d\n",
      glb->totnum);
   putlog(glb,str);
   sprintf(str,"Total errors: %7d\n",
      glb->err);
   putlog(glb,str);
   if (glb->replay)
      putlog(glb,"\n");
   } /* play */

void run(glbfmt *glb)
   {
   int wait;
   while (glb->run)
      {
      bldscr(glb);
      escape(glb);
      if (!glb->run) break;
      if (glb->run && glb->reset)
	 {
	 wait = glb->wait;
	 cls(glb);
	 clear();
	 bldmenu(glb);
	 free(glb->sd);
	 initglb(glb);
	 /* glb->reset  is set to zero */
	 /* glb->replay is set to zero */
         bldscr(glb);
	 glb->wait = wait;
	 play(glb);
	 } /* if reset */
      else if (glb->run && glb->replay)
	 {
	 wait = glb->wait;
	 initreplay(glb);
	 clear();
	 bldmenu(glb);
	 /* glb->reset  is set to zero */
	 /* glb->replay is set to zero */
         bldscr(glb);
	 glb->wait = wait;
	 play(glb);
	 } /* if replay same game */
      else
	 {
	 play(glb);
	 } /* not (reset or replay) */
      if (glb->run && !glb->reset) slp(glb);
      else if (glb->run && !glb->replay) slp(glb);
      } /* display iteration */
   } /* run */

int main(int argc,char **argv)
   {
   glbfmt *glb;
   if (argc != 1) putstx(*argv);
   glb = (glbfmt *) malloc(sizeof(glbfmt));
   if (glb == NULL)
      {
      fprintf(stderr,"sudoku: out of memory "
	 "allocating glb\n");
      exit(1);
      } /* if out of mem */
   initnc();
   initglb(glb);
   bldmenu(glb);
   bldscr(glb);
   run(glb);
   endwin();
   cls(glb);
   free(glb->sd);
   free(glb);
   return(0);
   } /* main */
